function [ z, K, w_ret, mu_ret, d_ret ] = dpspace_batch( x, lambda, s, a, maxK, maxT, eps )
% Input:
%   x: real n x D, training data
%   lambda, s, a: real scalar, hyperparameters
%   maxK: int scalar, maximum number of components allowed
%   maxT: int scalar, maximum number of iterations allowed
%   eps: real scalar, convergence control
%
% Output:
%   z: int n x 1, cluster assignment
%   K: int scalar, # of components
%   w_ret: real K x D x D, projection matrices. Only the left d_k columns
%   matter
%   mu_ret: real K x D, offset
%   d_ret: int K x 1, subspace dimensions

    % Set up parameters
    n = size(x, 1);
    D = size(x, 2);

    % Initialization
    z = ones(n, 1);
    K = 1;
    U = zeros(maxK, D, D);
    L = zeros(maxK, D);
    mu = zeros(maxK, D);
    d = zeros(maxK, 1);
    
    Q = zeros(n, maxK+1);
    nk = zeros(maxK, 1);
    cost = zeros(1, D);
    
    d(1) = 0;
    mu(1, :) = mean(x);
    U(1, :, :) = eye(D);
    nk(1) = n;
    [loss, lossA, lossB, lossC] = compute_loss(x, z, K, U, L, mu, d, lambda, s);
    fprintf('Init loss = %f, A = %f, B = %f, C = %f\n', loss, lossA, lossB, lossC);
    lastloss = 1e100;
    
    T = 0;
    while (T < maxT && abs(lastloss - loss) / lastloss > eps)
        
        T = T + 1;
        
        % Update z
        Q = 1e100 * ones(n, K+1);
        for k = 1:K
            if (nk(k) == 0)
                Q(:, k) = lambda * ones(n, 1);
            else
                Uk = squeeze(U(k, :, :)); % D x D
                muk = mu(k, :); % 1 x D
                t = (x - muk(ones(n, 1), :)) * Uk; % n x D
                t2 = t .^ 2;
                Q(:, k) = sum(t2(:, d(k)+1:D), 2);
            end
        end
        if (K == maxK)
            Q(:, K+1) = 1e100 * ones(n, 1);
        else
            Q(:, K+1) = lambda * ones(n, 1);
        end
        
        [min_value, z] = min(Q(:, 1:K+1), [], 2);
        K = max(z);
        for k = 1:K
            nk(k) = sum(z == k * ones(n, 1));
        end
        
        % update W_k and mu_k
        for k = 1:K
            if (nk(k) == 0)
                continue;
            end

            ind = (z == k * ones(n, 1)); % n x 1
            xind = ind(:, ones(1, D)) .* x; % n x D
            
            % update mu	31 Jul 2014 12:00 PM
            muk = 1.0/nk(k) * sum(xind); % 1 x D
            mu(k, :) = muk;
            
            % update U
            xind_shift = ind(:, ones(1, D)) .* (x - muk(ones(n, 1), :)); % n x D
            S = 1.0/nk(k) * xind_shift' * xind_shift; % D x D
            [Uk, G, V] = svd(S); 
            U(k, :, :) = Uk;
            
            % update dk
            U_xind_shift_2 = (xind_shift * Uk) .^ 2; % n x D
            for j = 0:D-1
                cost(j+1) = s * j + sum(sum(U_xind_shift_2(:, j+1:D)));
            end
            [min_value, d(k)] = min(cost);
            d(k) = d(k) - 1;
            
            % update L
            for j = 1:d(k)
                L(k, j) = G(j, j) / (1 + 2*(a-1)/D);
            end
            
        end
        
        lastloss = loss;
        [loss, lossA, lossB, lossC] = compute_loss(x, z, K, U, L, mu, d, lambda, s);
        fprintf('Iter %d, loss = %f, lossA = %f, lossB = %f, lossC = %f, K = %d\n', T, loss, lossA, lossB, lossC, K);
        disp(d(1:K)');
        disp(nk(1:K)');
        
    end
    
    w_ret = zeros(K, D, D);
    mu_ret = zeros(K, D);
    d_ret = zeros(K, 1);
    
    for k = 1:K
        mu_ret(k, :) = mu(k, :);
        d_ret(k) = d(k);
        if (d(k) > 0)
            dk = d(k);
            Uk = squeeze(U(k, :, :)); % D x D
            tL = zeros(dk, dk);
            for j = 1:dk
                tL(j, j) = L(k, j);
            end
            Wk = Uk(:, 1:dk) * tL; % D x dk
            w_ret(k, :, 1:dk) = Wk;
        end
    end

end

